# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2025.
# This code is licensed under the MIT License.

"""This file provides some utility functions for calling the ``isqc`` compiler
from python."""

from __future__ import annotations

import json
import logging
import os
import re
import shutil
import subprocess
from collections.abc import Sequence
from enum import Enum
from pathlib import Path

from .isqc_path import _IsqcPath, get_isqc_path, set_isqc_path

_default_isqc_path = _IsqcPath()


class IsqcError(Exception):
    """IsQ compiler Error."""


class CompileTarget(Enum):
    QIR = "qir"
    OPEN_QASM3 = "open-qasm3"
    QCIS = "qcis"

    @classmethod
    def from_string(cls, value: str) -> "CompileTarget":
        """Convert a string to the corresponding CompileTarget enum value."""
        value = value.lower()  # Case-insensitive comparison
        for target in cls:
            if target.value == value:
                return target
        raise ValueError(
            f"Invalid target: {value}. Valid options are: {[t.value for t in cls]}"
        )


def isqc_compile(
    file: str,
    target: CompileTarget | str = CompileTarget.QIR,
    int_param: Sequence[int] | int | None = None,
    double_param: Sequence[float] | float | None = None,
    additional_args: str = "",
) -> None:
    """This function encapsulates the ``compile`` of isQ compiler.

    Args:
        file: The path to the file that needs to be compiled.
        target: The compiled target output form (either a CompileTarget enum or a string).
        int_param: An integer variable (array) passed in when compiling.
        double_param: An double variable (array) passed in when compiling.
        additional_args: Other arguments passed in when compiling, see more:
                         https://www.arclightquantum.com/isq-docs/latest/

    """

    if isinstance(target, str):
        target = CompileTarget.from_string(target)
    if target == CompileTarget.OPEN_QASM3:
        raise IsqcError("open-qasm3 is broken")

    file = str(Path(file).expanduser())
    int_cmd, double_cmd = _deal_params(int_param, double_param)
    isqc_path = str(Path(_default_isqc_path.default_isqc_path) / "isqc")

    compile_cmd = [isqc_path, "compile", file, "--target", target.value]

    if int_cmd:
        compile_cmd.extend(int_cmd.split())
    if double_cmd:
        compile_cmd.extend(double_cmd.split())
    if additional_args:
        compile_cmd.extend(additional_args.split())

    out = subprocess.run(
        compile_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,  # Decode output as text
    )

    if out.returncode != 0:
        raise IsqcError(
            f"Compile Failed! "
            f"Error code: {out.returncode}. "
            f"Error message: {out.stdout} {out.stderr}"
            # isqc stdout and stderr
        )


def isqc_simulate(
    file: str | Path,
    shots: int = 100,
    int_param: Sequence[int] | int | None = None,
    double_param: Sequence[float] | float | None = None,
    debug: bool = False,
    additional_args: str = "",
) -> tuple[dict[str, int], str, int] | tuple[str, str, int]:
    """This function encapsulates the ``simulate`` of isQ compiler.

    Args:
        file: The path to the file that needs to be compiled.
        shots: Shots number of quantum simulation.
        int_param: An integer variable (array) passed in when compiling.
        double_param: An double variable (array) passed in when compiling.
        additional_args: Other arguments passed in when compiling, see more:
                         https://www.arclightquantum.com/isq-docs/latest/

    Returns:
        A dictionary with simulation results if successful, otherwise a tuple containing
        stdout, stderr, and the return code.

    Raises:
        IsqcError: If the file format is unsupported or the file does not exist.
        subprocess.CalledProcessError: If the simulation command fails.

    """

    if isinstance(file, str):
        file = Path(file).expanduser()
    if file.suffix == ".so":
        so_file = file
    elif file.suffix == ".isq":
        so_file = file.with_suffix(".so")
    else:
        raise IsqcError(
            f"`{file}`'s format is not supported. Please use `.isq` or `.so`"
        )

    if not so_file.exists():
        raise IsqcError(f"`{so_file}` does not exist. Please compile isQ file first.")

    int_cmd, double_cmd = _deal_params(int_param, double_param)
    isqc_path = os.path.join(_default_isqc_path.default_isqc_path, "isqc")
    simulate_cmd = [isqc_path, "simulate", str(so_file), "--shots", str(shots)]

    if int_cmd:
        simulate_cmd.extend(int_cmd.split())
    if double_cmd:
        simulate_cmd.extend(double_cmd.split())
    if additional_args:
        simulate_cmd.extend(additional_args.split())
    if debug:
        simulate_cmd.append("--debug")

    result = subprocess.run(
        simulate_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
        text=True,
    )
    if result.returncode == 0:
        return dict(json.loads(result.stdout)), result.stderr, result.returncode
    else:
        return result.stdout, result.stderr, result.returncode


def isqc_run(
    file: str,
    target: CompileTarget = CompileTarget.QIR,
    shots: int = 100,
    int_param: Sequence[int] | int | None = None,
    double_param: Sequence[float] | float | None = None,
) -> tuple[dict[str, int], str, int] | tuple[str, str, int]:
    """This method executes ``compile`` and ``simulate`` in sequence."""

    isqc_compile(file=file, target=target)

    return isqc_simulate(
        file=file, shots=shots, int_param=int_param, double_param=double_param
    )


def _deal_params(
    int_param: Sequence[int] | int | None = None,
    double_param: Sequence[float] | float | None = None,
) -> tuple[str, str]:
    """Convert integer and double parameters into command-line arguments for the isQ compiler.

    Args:
        int_param: A single integer, a sequence of integers, or None.
        double_param: A single float, a sequence of floats, or None.

    Returns:
        A tuple containing the command-line arguments for integer parameters and double parameters.
    """

    int_cmd = _format_params(int_param, "-i")
    double_cmd = _format_params(double_param, "-d")

    return int_cmd, double_cmd


def _format_params(param, flag):
    if param is None:
        return ""
    if isinstance(param, (int, float)):
        param = [param]
    return " ".join(f"{flag} {value}" for value in param)


def gen_qcis_from_so(
    file: str | Path,
    int_param: Sequence[int] | int | None = None,
    double_param: Sequence[float] | float | None = None,
    additional_args: str = "",
) -> None:
    """According to the compilation plan of isQ compiler, this function is not
    open to users.
    """

    if isinstance(file, str):
        file = Path(file).expanduser()

    int_cmd, double_cmd = _deal_params(int_param, double_param)

    qcis_path = file.with_suffix(".qcis")

    FindSimulator.make_simulator_bin()
    simulator_exec = FindSimulator.get_simulator_bin_path()

    cmd = [
        simulator_exec,
        str(file),
        "--qcisgen",
        "-e",
        "__isq__entry",
    ]

    if int_cmd:
        cmd.extend(int_cmd.split())
    if double_cmd:
        cmd.extend(double_cmd.split())
    if additional_args:
        cmd.extend(additional_args.split())

    with open(qcis_path, "w") as output_file:
        out = subprocess.run(
            cmd,
            stdout=output_file,
            stderr=output_file,
            text=True,
        )

    if out.returncode != 0:
        raise IsqcError(
            f"Compile Failed! "
            f"Error code: {out.returncode}. "
            f"Error message: {out.stdout} {out.stderr}"
            # isqc stdout and stderr
        )


class FindSimulator:
    """To build a SIMULATOR BIN for isqc."""

    def __init__(self) -> None:
        self.make_simulator_bin()

    @staticmethod
    def set_simulator_file_name():
        return "SIMULATOR"

    @staticmethod
    def get_dir_name_of_simulator(isqcDIR):
        storeDIR = os.path.join(isqcDIR, "nix", "store")
        messList = os.listdir(storeDIR)
        for line in messList:
            if "simulator" in line:
                return line.strip()
        logging.error("cannot find simulator dir")

    @staticmethod
    def get_isQ_dir():
        if not get_isqc_path():
            isqcBIN = shutil.which("isqc")
            if isqcBIN:
                isqcDIR = os.path.abspath(os.path.dirname(isqcBIN))
                return isqcDIR
            return None
        return get_isqc_path()

    @classmethod
    def get_simulator_bin_path(cls):
        return os.path.join(cls.get_isQ_dir(), cls.set_simulator_file_name())

    @classmethod
    def make_simulator_bin(cls):
        """
        Automatically find the directory of isqc, create a SIMULATOR file
        with appropriate content, and set file permissions.
        """
        simBinFile = os.path.abspath(cls.get_simulator_bin_path())

        if os.path.exists(simBinFile):
            print(f"{simBinFile} already exists.")
            return

        isQdir = os.path.abspath(cls.get_isQ_dir())
        if not os.path.isdir(isQdir):
            print(f"Error: {isQdir} is not a valid directory.")
            return
        messName = cls.get_dir_name_of_simulator(isQdir)
        simulator_path = os.path.join("/nix/store", messName, "bin", "simulator")
        isqcFile = os.path.join(isQdir, "isqc")
        if not os.path.isfile(isqcFile):
            print(f"Error: {isqcFile} does not exist or is not a file.")
            return
        try:
            with open(isqcFile, "r") as f:
                isqcContent = f.read()
            slides = isqcContent.split(" ")
            if len(slides) < 2:
                print("Error: The isqc file format is not valid.")
                return
            slides[-2] = simulator_path  # 更新模拟器路径
            simBINcontent = " ".join(slides)
            with open(simBinFile, "w") as f:
                f.write(simBINcontent)
            os.chmod(simBinFile, 0o555)
            print(f"Simulator binary created successfully at {simBinFile}.")

        except OSError as e:
            print(f"Failed to create simulator binary: {e}")



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""
A Neural Network abstract class for all.
This part code is inspired by IBM qiskit.
For more information, check out:https://qiskit.org/
"""

from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np

try:
    from sparse import SparseArray
except Exception:

    class SparseArray:
        """Empty SparseArray class
        Replacement if sparse.SparseArray is not present.
        """


class NeuralNetworkError(Exception):
    """Neural network error."""


class NeuralNetwork(ABC):
    """Abstract Neural Network class providing forward and backward pass and
    handling batched inputs. This is to be implemented by other (quantum)
    neural networks.
    """

    def __init__(
        self,
        num_inputs: int,
        num_weights: int,
        sparse: bool,
        output_shape: int | tuple[int, ...],
        input_gradients: bool = True,
    ) -> None:
        """Initialize quantum neural network.

        Args:
            num_inputs: The number of input features.
            num_weights: The number of trainable weights.
            sparse: Determines whether the output is a sparse array or not.
            output_shape: The shape of the output.
            input_gradients: Determines whether to compute gradients with
                             respect to input data.

        Raises:
            NeuralNetworkError: Invalid parameter values.

        """
        if num_inputs < 0:
            raise NeuralNetworkError(
                f"Number of inputs cannot be negative: {num_inputs}!",
            )

        self._num_inputs = num_inputs

        if num_weights < 0:
            raise NeuralNetworkError(
                f"Number of weights cannot be negative: {num_weights}!"
            )

        self._num_weights = num_weights

        self._sparse = sparse

        # output shape may be derived later, so check it only if it is not None
        if output_shape is not None:
            self._output_shape = self._validate_output_shape(output_shape)

        self._input_gradients = input_gradients

    @property
    def num_inputs(self) -> int:
        """Returns the number of input features."""
        return self._num_inputs

    @property
    def num_weights(self) -> int:
        """Returns the number of trainable weights."""
        return self._num_weights

    @property
    def sparse(self) -> bool:
        """Returns whether the output is sparse or not."""
        return self._sparse

    @property
    def output_shape(self) -> tuple[int, ...]:
        """Returns the output shape."""
        return self._output_shape

    @property
    def input_gradients(self) -> bool:
        """Returns whether gradients with respect to input data are computed by
        this neural network in the ``backward`` method or not. By default such
        gradients are not computed."""
        return self._input_gradients

    @input_gradients.setter
    def input_gradients(self, input_gradients: bool) -> None:
        """Turn on/off computation of gradients with respect to input data."""
        self._input_gradients = input_gradients

    def _validate_output_shape(self, output_shape):
        if isinstance(output_shape, int):
            output_shape = (output_shape,)
        if not np.all([s > 0 for s in output_shape]):
            raise NeuralNetworkError(
                f"Invalid output shape, all components must be > 0, "
                f"but got: {output_shape}."
            )
        return output_shape

    def _validate_input(
        self, input_data: float | list[float] | np.ndarray | None
    ) -> tuple[np.ndarray | None, tuple[int, ...] | None]:
        if input_data is None:
            return None, None
        input_ = np.array(input_data)
        shape = input_.shape
        if len(shape) == 0:
            # there's a single value in the input.
            input_ = input_.reshape((1, 1))
            return input_, shape

        if shape[-1] != self._num_inputs:
            raise NeuralNetworkError(
                f"Input data has incorrect shape, last dimension "
                f"is not equal to the number of inputs: "
                f"{self._num_inputs}, but got: {shape[-1]}."
            )

        if len(shape) == 1:
            # add an empty dimension for samples (batch dimension)
            input_ = input_.reshape((1, -1))
        elif len(shape) > 2:
            # flatten lower dimensions, keep num_inputs as a last dimension
            input_ = input_.reshape((np.product(input_.shape[:-1]), -1))

        return input_, shape

    def _preprocess_forward(
        self,
        input_data: np.ndarray | None,
        weights: np.ndarray | None,
    ) -> tuple[np.ndarray | None, int | None]:
        """Pre-processing during forward pass of the network for the
        primitive-based networks.
        """
        if input_data is not None:
            num_samples = input_data.shape[0]
            if weights is not None:
                weights = np.broadcast_to(weights, (num_samples, len(weights)))
                parameters = np.concatenate((input_data, weights), axis=1)
            else:
                parameters = input_data
        else:
            if weights is not None:
                num_samples = 1
                parameters = np.broadcast_to(
                    weights,
                    (
                        num_samples,
                        len(weights),
                    ),
                )
            else:
                return None, None
        return parameters, num_samples

    def _validate_weights(
        self, weights: float | list[float] | np.ndarray | None
    ) -> np.ndarray | None:
        if weights is None:
            return None
        weights_ = np.array(weights)
        return weights_.reshape(self._num_weights)

    def _validate_forward_output(
        self, output_data: np.ndarray, original_shape: tuple[int, ...]
    ) -> np.ndarray:
        if original_shape and len(original_shape) >= 2:
            output_data = output_data.reshape(
                (*original_shape[:-1], *self._output_shape)
            )

        return output_data

    def _validate_backward_output(
        self,
        input_grad: np.ndarray,
        weight_grad: np.ndarray,
        original_shape: tuple[int, ...],
    ) -> tuple[np.ndarray | SparseArray, np.ndarray | SparseArray]:
        if input_grad is not None and np.prod(input_grad.shape) == 0:
            input_grad = None

        if input_grad is not None and original_shape and len(original_shape) >= 2:
            input_grad = input_grad.reshape(
                (*original_shape[:-1], *self._output_shape, self._num_inputs)
            )

        if weight_grad is not None and np.prod(weight_grad.shape) == 0:
            weight_grad = None

        if weight_grad is not None and original_shape and len(original_shape) >= 2:
            weight_grad = weight_grad.reshape(
                (*original_shape[:-1], *self._output_shape, self._num_weights)
            )

        return input_grad, weight_grad

    def forward(
        self,
        input_data: float | list[float] | np.ndarray | None,
        weights: float | list[float] | np.ndarray | None,
    ) -> np.ndarray | SparseArray:
        """Forward pass of the network.

        Args:
            input_data: input data of the shape (num_inputs). In case of a
                single scalar input it is directly cast to and interpreted like
                a one-element array.
            weights: trainable weights of the shape (num_weights). In case of a
                single scalar weightit is directly cast to and interpreted like
                a one-element array.

        Returns:
            The result of the neural network of the shape (output_shape).

        """
        input_, shape = self._validate_input(input_data)
        weights_ = self._validate_weights(weights)
        output_data = self._forward(input_, weights_)
        return self._validate_forward_output(output_data, shape)

    @abstractmethod
    def _forward(
        self, input_data: np.ndarray | None, weights: np.ndarray | None
    ) -> np.ndarray | SparseArray:
        """Forward calculation."""

    def backward(
        self,
        input_data: float | list[float] | np.ndarray | None,
        weights: float | list[float] | np.ndarray | None,
    ) -> tuple[np.ndarray | SparseArray | None, np.ndarray | SparseArray | None]:
        input_, shape = self._validate_input(input_data)
        weights_ = self._validate_weights(weights)
        input_grad, weight_grad = self._backward(input_, weights_)
        (
            input_grad_reshaped,
            weight_grad_reshaped,
        ) = self._validate_backward_output(input_grad, weight_grad, shape)

        return input_grad_reshaped, weight_grad_reshaped

    @abstractmethod
    def _backward(
        self, input_data: np.ndarray | None, weights: np.ndarray | None
    ) -> tuple[np.ndarray | SparseArray | None, np.ndarray | SparseArray | None]:
        """Backward calculation."""



import tempfile
from pathlib import Path

import pytest

from isqtools.utils import CompileTarget, IsqcError, isqc_compile

TEST_FILE_CONTENT = """
import std;

qbit q[2];

procedure main() {
    H(q[0]);
    H(q[1]);
    M(q[0]);
    M(q[1]);
}
"""


@pytest.mark.parametrize("target", [CompileTarget.QIR, "qir"])
def test_compile_qir(target):
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(TEST_FILE_CONTENT)
        isqc_compile(file=str(temp_file_path), target=target)
        output_file = temp_file_path.with_suffix(".so")
        assert output_file.exists()


@pytest.mark.parametrize("target", [CompileTarget.QCIS, "qcis"])
def test_compile_qcis(target):
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(TEST_FILE_CONTENT)
        isqc_compile(file=str(temp_file_path), target=target)
        output_file_so = temp_file_path.with_suffix(".so")
        output_file_qcis = temp_file_path.with_suffix(".qcis")
        assert output_file_so.exists() and output_file_qcis


@pytest.mark.parametrize("target", [CompileTarget.OPEN_QASM3, "open-qasm3"])
def test_compile_openqasm3(target):
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(TEST_FILE_CONTENT)
        with pytest.raises(IsqcError, match="open-qasm3 is broken"):
            isqc_compile(file=str(temp_file_path), target=target)


def test_compile_with_invalid_file():
    """Test the compile function with an invalid file path."""
    invalid_file_path = "/invalid/path/to/file.isq"

    with pytest.raises(IsqcError):
        isqc_compile(file=invalid_file_path)


def test_compile_with_empty_file():
    """Test the compile function with an empty file."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w"):
            pass
        with pytest.raises(IsqcError):
            isqc_compile(file=str(temp_file_path))



import tempfile
from pathlib import Path

import pytest

from isqtools import IsqCircuit
from isqtools.backend import TianyBackend

pytest.importorskip("cqlib")
TEST_FILE_CONTENT = """
import std;
procedure main(int seq[], double double_[]) {
    int qbit_topo[]=[0,1,2,3,4,5,6];
    qbit q[seq[0]];
    X(q[qbit_topo[0]]);
    M(q[qbit_topo[0]]);
}
"""


def test_tianyan_simulator():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_file_path = temp_dir_path / "temp_file.isq"
        with open(temp_file_path, "w") as temp_file:
            temp_file.write(TEST_FILE_CONTENT)

        backend = TianyBackend(
            token="Ev5Ex6zUru41Grtoixg68ac2fCPW5cBrWxLVyPkavG0=",
            machine_name="tianyan_sw",
            max_wait_time=1,
        )

        qc = IsqCircuit(
            file=str(temp_file_path),
            sample=True,
            shots=100,
            backend=backend,
            int_param=[1],
        )
        result = qc.measure()

        query_id_single = backend.query_id_single
        while not result:
            result = backend.query_sample_result(
                query_id_single=query_id_single, max_wait_time=1
            )



# This code is part of isQ.
# (C) Copyright ArcLight Quantum 2023.
# This code is licensed under the MIT License.

"""Qcis2 hardware backend.
For more information, check out: https://quantumcomputer.ac.cn/
"""

from __future__ import annotations

import logging
import secrets
import string
import time

from numpy import pi

from .hardware_backend import HardwareBackend, ISQTask, TaskState, TaskType

try:
    from pyezQ import Account

    # pyezQ is necessary for the execution of the hardware.
except ImportError as e:
    logging.info(f"Failed to import pyezQ module: {e}")


def load_params(qcis: str, **kw) -> str:
    """Use the periodicity to convert the angles in all rotation operations to
    the range of -2pi to 2pi.

    Args:
        qcis: qcis for quantum circuits.

    Returns:
        qcis being converted.

    """
    line_data = qcis.split("\n")
    new_qcis = []
    for line in line_data:
        line = line.strip()
        if not line:
            continue
        str_arr = line.split(" ")
        if str_arr[0] in ["RX", "RY", "RZ", "RXY"]:
            theta_arr = []
            for v in str_arr[2:]:
                theta = float(eval(v, kw))
                theta %= 4 * pi
                if theta > 2 * pi:
                    theta -= 4 * pi
                theta_arr.append(str(theta))
            new_line = str_arr[:2] + theta_arr
            new_qcis.append(" ".join(new_line))
        else:
            new_qcis.append(line)
    return "\n".join(new_qcis)


def split_rotation_gates(qcis: str) -> str:
    """The qcis hardware currently has certain limitations. For the rotation
    operation, when the angle is in the range of -2pi to -pi and pi to 2pi, it
    is necessary to divide the angle into two and convert it into two rotation
    operations.

    Args:
        qcis: qcis for quantum circuits.

    Returns:
        qcis being converted.

    """
    ir_list = qcis.split("\n")
    ir_list_copy = ir_list.copy()
    for i, ir in enumerate(ir_list):
        if ir.startswith("RXY"):
            ir_list_copy_list = ir_list_copy[i].split()
            if float(ir_list_copy_list[3]) > pi or float(ir_list_copy_list[3]) < -pi:
                q, phi, theta = (
                    ir_list_copy_list[1],
                    float(ir_list_copy_list[2]),
                    float(ir_list_copy_list[3]),
                )
                decompose = []
                phi1 = pi / 2 - phi
                if abs(phi1) > pi:
                    decompose.append(f"RZ {q} {phi1/2}")
                    decompose.append(f"RZ {q} {phi1/2}")
                else:
                    decompose.append(f"RZ {q} {phi1}")
                decompose.append(f"X2P {q}")
                decompose.append(f"RZ {q} {theta / 2}")
                decompose.append(f"RZ {q} {theta / 2}")
                decompose.append(f"X2M {q}")
                phi2 = phi - pi / 2
                if abs(phi2) > pi:
                    decompose.append(f"RZ {q} {phi2/2}")
                    decompose.append(f"RZ {q} {phi2/2}")
                else:
                    decompose.append(f"RZ {q} {phi2}")
                ir_list_copy[i] = "\n".join(decompose)
        elif any([ir.startswith("RX"), ir.startswith("RY"), ir.startswith("RZ")]):
            ir_list_copy_list = ir_list_copy[i].split()
            if float(ir_list_copy_list[2]) > pi or float(ir_list_copy_list[2]) < -pi:
                ir_list_copy_list_new = ir_list_copy_list.copy()
                ir_list_copy_list_new[2] = str(float(ir_list_copy_list[2]) / 2)
                ir_list_copy_new = " ".join(ir_list_copy_list_new)
                new_str = ir_list_copy_new + "\n" + ir_list_copy_new
                ir_list_copy[i] = new_str

    return "\n".join(ir_list_copy)


def get_rand_str() -> str:
    """Generate a cryptographically secure random string."""
    rand_str = "".join(secrets.choice(string.ascii_lowercase) for _ in range(10))
    return rand_str


class QcisBackend2Error(Exception):
    """Qcis backend error"""


class QcisBackend2(HardwareBackend):
    """the backend of qcis quantum hardware"""

    def __init__(
        self,
        login_key: str,
        machine_name: str,
        lab_id: str | None = None,
        exp_name: str | None = None,
        # remark: str = "arclight",
        max_wait_time: int = 60,
        sleep_time: int = 3,
        run_time: int | None = None,
        mapping: bool = False,
        is_prob: bool = False,
        logger=None,
    ) -> None:
        if logger is None:
            logger = logging.getLogger(__name__)

        self.sleep_time = sleep_time
        self.run_time = run_time
        self.mapping = mapping
        self.is_prob = is_prob
        self.logger = logger

        self._account = Account(login_key=login_key, machine_name=machine_name)

        if lab_id:
            self.lab_id = lab_id
        else:
            if exp_name is None:
                time_now = time.strftime(
                    "%Y_%m_%d_%H_%M_%S", time.localtime(time.time())
                )
                exp_name = f"arclight_{time_now}"
            create_res = self._account.create_experiment(
                exp_name,
                # remark=remark,
                # NOTE: different from ezQgd
            )
            if create_res == 0:
                print("新建实验合集失败")
            else:
                print("新建实验合辑成功，ID=", create_res)
                self.lab_id = create_res

        self._qid = 0
        self._max_wait_time = max_wait_time

    def sample(
        self,
        data: str,
        shots: int = 100,
        **kwargs,
    ) -> dict[str, int]:
        """Call quantum hardware for sampling.

        Args:
            data: QCIS strings, This represents the information of the quantum
                  circuit.
            shots: Shots numbers.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            This method returns a dictionary, the key in the dictionary is
            the quantum state, and the value is the shots of the quantum state.
        """

        qcis_split_rotation_gates = split_rotation_gates(
            load_params(data, **kwargs),
        )
        if self.mapping:
            qcis_split_rotation_gates = self._account.qcis_mapping_isq(
                qcis_split_rotation_gates
            )
        time_now = time.strftime("%Y_%m_%d_%H_%M_%S", time.localtime(time.time()))
        query_id = self._account.submit_job(
            circuit=qcis_split_rotation_gates,
            lab_id=self.lab_id,
            version=f"{self._qid}_{time_now}",
            num_shots=shots,
        )
        self._qid += 1
        if query_id:
            self.logger.info("Qcis hardware: submitted the task successfully.")
            task = ISQTask(query_id, TaskType.QCIS2, TaskState.WAIT, self, shots)
        else:
            self.logger.error(
                "Qcis submission failed, please confirm whether the hardware"
                "is normal, or try again later."
            )
            task = ISQTask(0, TaskType.QCIS2, TaskState.FAIL, self, shots)

        if self.run_time is not None:
            start_time = time.time()

        while task.state == TaskState.WAIT:
            try:
                task.result()
                time.sleep(self.sleep_time)
            except Exception as e:
                continue

            # Wait for a period of time before doing an http request
            if self.run_time is not None:
                if time.time() > start_time + self.run_time:
                    break
        return task.result()
